// -*- C++ -*-
/*!
 * @file  Bumblebee2Camera.cpp * @brief Bumblebee2 Camera component * $Date$ 
 *
 * $Id$ 
 */
#include "Bumblebee2Camera.h"

// Module specification
// <rtc-template block="module_spec">
static const char* bumblebee2camera_spec[] =
  {
    "implementation_id", "Bumblebee2Camera",
    "type_name",         "Bumblebee2Camera",
    "description",       "Bumblebee2 Camera component",
    "version",           "1.0",
    "vendor",            "Tada,Mizoguchi Lab,Tokyo University of Science",
    "category",          "RobotModule",
    "activity_type",     "SPORADIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.WIDTH", "320",
    "conf.default.HEIGHT", "240",
    ""
  };
// </rtc-template>

Bumblebee2Camera::Bumblebee2Camera(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_ColorOut("Color", m_Color),
    m_ColorRightOut("ColorRight", m_ColorRight),
    m_DisparityOut("Disparity", m_Disparity),
    m_XYZLongOut("XYZLong", m_XYZLong),
    m_XYZDoubleOut("XYZDouble", m_XYZDouble),
    m_CapPORTPort("CapPORT")

    // </rtc-template>
{
/*
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers

  // Set OutPort buffer
  registerOutPort("Color", m_ColorOut);
  registerOutPort("ColorRight", m_ColorRightOut);
  registerOutPort("Disparity", m_DisparityOut);
  registerOutPort("XYZLong", m_XYZLongOut);
  registerOutPort("XYZDouble", m_XYZDoubleOut);

  // Set service provider to Ports
  m_CapPORTPort.registerProvider("CameraInfo", "CameraInfo", m_CameraInfo);

  // Set service consumers to Ports

  // Set CORBA Service Ports
  registerPort(m_CapPORTPort);
*/
  // </rtc-template>

}

Bumblebee2Camera::~Bumblebee2Camera()
{
}


RTC::ReturnCode_t Bumblebee2Camera::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers

  // Set OutPort buffer
  addOutPort("Color", m_ColorOut);
  addOutPort("ColorRight", m_ColorRightOut);
  addOutPort("Disparity", m_DisparityOut);
  addOutPort("XYZLong", m_XYZLongOut);
  addOutPort("XYZDouble", m_XYZDoubleOut);

  // Set service provider to Ports
  m_CapPORTPort.registerProvider("CameraInfo", "CameraInfo", m_CameraInfo);

  // Set service consumers to Ports

  // Set CORBA Service Ports
  addPort(m_CapPORTPort);

  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("WIDTH", m_WIDTH, "320");
  bindParameter("HEIGHT", m_HEIGHT, "240");

	cout<<"Bumblebee2 Component"<<endl;
  // </rtc-template>
  return RTC::RTC_OK;
}



RTC::ReturnCode_t Bumblebee2Camera::onFinalize()
{
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t Bumblebee2Camera::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t Bumblebee2Camera::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t Bumblebee2Camera::onActivated(RTC::UniqueId ec_id)
{
	static double xx[320 * 240];
	static double yy[320 * 240];
	static double zz[320 * 240];
	static int aa[320 * 240];

	m_CameraInfo.channels = 3;
	m_CameraInfo.w = m_WIDTH;
	m_CameraInfo.h = m_HEIGHT;


    if(bumblebee.Initialize() == 1){
		cout<<"Error"<<endl;	
		//return RTC::RTC_ERROR;
    }

    CapHeight = bumblebee.getHeight();
    CapWidth = bumblebee.getWidth();	
    CapBytesPerPixel = bumblebee.getChannels();
    bColor = bumblebee.getColor();

	Height = m_HEIGHT;
	Width = m_WIDTH;
	Channels = 3;

	bumblebee.triclopsInit();

    disparity 	= new short int [ Height * Width ];
    xxx 	= new double[ Height * Width ];
    yyy 	= new double[ Height * Width ];
    zzz 	= new double[ Height * Width ];


	//モノクロ画像（おそらくほとんど使わない）
    leftmono 	= cvCreateImage(cvSize(CapWidth, CapHeight), IPL_DEPTH_8U, 1); 
    rightmono 	= cvCreateImage(cvSize(CapWidth, CapHeight), IPL_DEPTH_8U, 1);
    rightmonoResize = cvCreateImage(cvSize(Width, Height), IPL_DEPTH_8U, 1);
    leftmonoResize 	= cvCreateImage(cvSize(Width, Height), IPL_DEPTH_8U, 1);

	//Bumblebeeから生画像を取得
    leftRGB 	= cvCreateImage(cvSize(CapWidth, CapHeight), IPL_DEPTH_8U, Channels); 
    rightRGB 	= cvCreateImage(cvSize(CapWidth, CapHeight), IPL_DEPTH_8U, Channels);

	//取得した画像を指定したサイズにリサイズ
    rightResize = cvCreateImage(cvSize(Width, Height), IPL_DEPTH_8U, Channels);
    leftResize 	= cvCreateImage(cvSize(Width, Height), IPL_DEPTH_8U, Channels);

	//RGBの順番をBGRに変換する
    left 		= cvCreateImage(cvSize(Width, Height), IPL_DEPTH_8U, Channels); 
    right 		= cvCreateImage(cvSize(Width, Height), IPL_DEPTH_8U, Channels);

	//キャリブレーションにより歪みを補正した画像を格納(右画像のみ)
	rightRect 	= cvCreateImage(cvSize(Width, Height), IPL_DEPTH_8U, Channels);

	//視差画像用	
    dispResize = cvCreateImage(cvSize(STEREO_WIDTH, STEREO_HEIGHT), IPL_DEPTH_8U, 1);
    disp       = cvCreateImage(cvSize(Width, Height), IPL_DEPTH_8U, 1);
	rect       = cvCreateImage(cvSize(Width, Height), IPL_DEPTH_8U, 1);	

    cout<<"Start"<<endl;
    cout<<"SIZE:"<<WIDTH<<":"<<HEIGHT<<endl;

	cvNamedWindow("Left image", CV_WINDOW_AUTOSIZE);
	cvNamedWindow("Right image", CV_WINDOW_AUTOSIZE);
	cvNamedWindow("Disparity image", CV_WINDOW_AUTOSIZE);


	frame = 0;
	sumtime = 0.0;
	t = 0.0;

  return RTC::RTC_OK;
}


RTC::ReturnCode_t Bumblebee2Camera::onDeactivated(RTC::UniqueId ec_id)
{
	delete[] xxx;
	delete[] yyy;
	delete[] zzz;
	delete[] disparity;

	cvReleaseImage(&left);
	cvReleaseImage(&leftRGB);
	cvReleaseImage(&leftResize);
	cvReleaseImage(&leftmonoResize);
	cvReleaseImage(&leftmono);
	cvReleaseImage(&right);
	cvReleaseImage(&rightRGB);
	cvReleaseImage(&rightResize);
	cvReleaseImage(&rightmono);
	cvReleaseImage(&rightmonoResize);
	cvReleaseImage(&rightRect);
	cvReleaseImage(&disp);
	cvReleaseImage(&dispResize);
	cvReleaseImage(&rect);
	//cvReleaseImage(&center);
	//cvReleaseImage(&centerRGB);
	//cvReleaseImage(&centermono);

	cvDestroyAllWindows();

	bumblebee.closeB();		//BumbleBee2の終了処理を最後にしないとまずいっぽい
	
	return RTC::RTC_OK;
}


RTC::ReturnCode_t Bumblebee2Camera::onExecute(RTC::UniqueId ec_id)
{
  	old_t = bumblebee.timer_ms();	//一周期分の時間を取得し始める

	short int key = cvWaitKey(10);
	//cout<<CapHeight<<":"<<CapWidth<<endl;
	if ( bColor )
	{
		bumblebee.getColorImage(leftRGB, rightRGB );
		cvResize(leftRGB, leftResize, CV_INTER_NN);
		cvCvtColor(leftResize, left, CV_RGB2BGR);	
		cvResize(rightRGB, rightResize, CV_INTER_NN);
		cvCvtColor(rightResize, rightRect, CV_RGB2BGR);	
		bumblebee.Calibration(rightRect, right);					
		cvShowImage("Left image", left);
		cvShowImage("Right image", right);
	}	
	else
	{
		bumblebee.getMonoImage(leftmono, rightmono );
		cvResize(leftRGB, leftmonoResize, CV_INTER_NN);
		cvResize(rightRGB, rightmonoResize, CV_INTER_NN);
		cvShowImage("Left image", leftmonoResize);
		cvShowImage("Right image", rightmonoResize);
	}		

	bumblebee.triclopsProcess();
	bumblebee.triclopsGetDisp(disparity);
	bumblebee.triclopsGetPoint(xxx, yyy, zzz);
	//bumblebee.triclopsRect(rect);
	for(int i = 0; i < Width * Height; i++){
		//dispResize->imageData[i]=zzz[i]*1000.0/2000.0*255.0;
		dispResize->imageData[i] = (double)disparity[i] / 16.0;
	}
	cvResize(dispResize, disp, CV_INTER_NN);
	cvShowImage("Disparity image", disp);



	m_Color.data.length(left->imageSize);
	memcpy((void *)&(m_Color.data[0]), left->imageData, left->imageSize);
	m_ColorOut.write();

	m_ColorRight.data.length(right->imageSize);
	memcpy((void *)&(m_ColorRight.data[0]), right->imageData, right->imageSize);
	m_ColorRightOut.write();

	m_Disparity.data.length(disp->imageSize);
	memcpy((void *)&(m_Disparity.data[0]), disp->imageData, disp->imageSize);
	m_DisparityOut.write();

	m_XYZLong.data.length(m_WIDTH * m_HEIGHT);
	for(int i = 0; i < m_WIDTH * m_HEIGHT; i++){
		m_XYZLong.data[i].x = (int)(xxx[i] * 1000.0);
		m_XYZLong.data[i].y = (int)(yyy[i] * 1000.0);
		m_XYZLong.data[i].z = (int)(zzz[i] * 1000.0);
	}
	m_XYZLongOut.write();

	m_XYZDouble.data.length(m_WIDTH * m_HEIGHT);
	for(int i = 0; i < m_WIDTH * m_HEIGHT; i++){
		m_XYZDouble.data[i].x = xxx[i];
		m_XYZDouble.data[i].y = yyy[i];
		m_XYZDouble.data[i].z = zzz[i];
	}
	m_XYZDoubleOut.write();


	now_t = bumblebee.timer_ms();
	t = (now_t - old_t);
	dt = t / 1000.0;
	sumtime += (t / 1000.0);

    switch( (char)key )
    {
		case 't':
				cout<<"time(ms): "<<sumtime<<" frame: "<<frame<<" fps: "<<(double)frame / sumtime<<endl;
			break;
		case 'v':
				//cvWriteFrame(vw_ms, meanshift2);
				//cvWriteFrame(vw_ms, spatiograms);
			break;
		case 'c':

			break;
		default:
	        break;
    }

	if(key == 27)
		onDeactivated(0);
	frame++;

	return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t Bumblebee2Camera::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t Bumblebee2Camera::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t Bumblebee2Camera::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t Bumblebee2Camera::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t Bumblebee2Camera::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


extern "C"
{
 
  void Bumblebee2CameraInit(RTC::Manager* manager)
  {
    coil::Properties profile(bumblebee2camera_spec);
    manager->registerFactory(profile,
                             RTC::Create<Bumblebee2Camera>,
                             RTC::Delete<Bumblebee2Camera>);
  }
  
};



